
	MODULE poisson_routines

	implicit none

!---------------------------------------------
! NOTE: USE ONLY ONE OF THE FOLLOWING INCLUDE
!       LINES BELOW
!---------------------------------------------
! for turing cluster AHPCC; NAG f95 compiler
! include "mpif90.h"
!---------------------------------------------
! for BLACKBEAR AHPCC; PGI mpif90 compiler
  include "mpif.h"
!---------------------------------------------

! Global Variables

	integer,  parameter		:: DP=kind(1.d0)
	real(DP), parameter		:: pi=3.141592653589793238462643383279502884197_dp

	integer,  parameter		:: ITMAX   = 100000	!stop if exceed this # of iterations
	integer,  parameter		:: NOUTPUT = 1000	!output after this # iterations

      	character (len=128)             :: filename_head = 'poisson.dat'
      	character (len=128)             :: filename_head_rho = 'rho.dat'

!  MPI Global Variables

	integer				:: grid_comm
	integer				:: grid_rank
	integer				:: coords(2)		! procr coords in grid_comm; P(i,j)
	integer				:: row_comm		! row sub communicator
	integer				:: col_comm		! col sub comminicator
	integer				:: ierror

	integer				:: np_rows		! # rows in the procr grid
	integer				:: np_cols		! # cols in the procr grid
	integer, allocatable		:: nmbr_local_rows(:)
	integer, allocatable		:: nmbr_local_cols(:)
	integer				:: N_local, M_local


! Things to Note:
! ---------------
! NOTE: MPI gives grid coordinates starting from index 0, i.e.
!       P(i,j) where 0 <= i <= np_rows-1, 0 <= j <= np_cols-1
!   	where np_rows is the # of rows in the processor grid
!       and   np_cols is the # of cols in the processor grid
!
! NOTE: the processor is designated by two indices. i.e P( coords(1), coords(2) )
!       these indices are obtained by a call to MPI_Cart_coords

	
	CONTAINS

!==========================================================================

      	function makerho_local(Lx, Ly, N, M)


!  initialize potential to a positive and negative guassian 
!  centered at (x=L/4,y=0) and (x=3/4*L,y=0), repsectively.

      	implicit none


	integer,	intent(in)	:: N,M			!global indices for whole grid
	real(DP),	intent(in)	:: Lx, Ly		!x and y lengths of grid
	real(DP)	  		:: makerho_local(N_local,M_local)

! Local Variables

	integer				:: rank, ierr
      	integer 			:: i,j
	integer				:: i_offset, j_offset
      	real(DP) 			:: a, x, y


! determine the global x and y index offsets corresponding to the grid coordinates
! of the processor. nmbr_local_rows() and nmbr_local_cols() are arrays of size
! np_rows and np_cols which contain the # row and # columns in the local matrices
! that each processor in the processor grid is reponsible for. Global indices
! are given by i_global = i_local + i_offset, j_global = j_local + j_offset 
! where 	1 <= i <= N,	1 <= i_local <= N_local
! and 		1 <= j <= M,	1 <= j_local <= M_local
!
! Note: here I use the usual fortran notation that i varies down columns
! 	and  j varies across rows. i is associated with x and j is associated with y


	i_offset = 0
	do i = 1, coords(1)
	   i_offset = i_offset + nmbr_local_rows(i)
	enddo
 		
	j_offset = 0
	do j = 1, coords(2)
	   j_offset = j_offset + nmbr_local_cols(j)
	enddo


! initialize width of gaussian

      	a = -16./Lx/Ly*log(0.1)

! create the double peaked gaussian
      
      	do j = j_offset + 1, j_offset + M_local
           y = (real(j)-.5)/M * Ly-Ly/2.
           do i = i_offset + 1, i_offset + N_local
              x = (real(i)-.5)/N*Lx
              makerho_local(i-i_offset,j-j_offset) = a/pi*( exp(-a*((x-Lx/4.)**2+y*y)) - &
              		                  		    exp(-a*((x-3*Lx/4.)**2+y*y)) )
           enddo
        enddo


      	end function makerho_local


!==========================================================================

	subroutine output( phi, filename_head, it )

	use string_utilities
	
	implicit none

	integer,		intent(in) 	:: it
	real(DP),		intent(in) 	:: phi(:,:)
	character(128),		intent(in) 	:: filename_head

!  Local Variables

	integer					:: i, j, N, M
	character(128)				:: filename

!  output file with name ' filename_head.*** ' where
!  *** is a 3 digit character string representation
!  of the integer i, with leading zeros, i.e. 1 -> 001, etc...

	N = size(phi,1); M = size(phi,2)
        filename = int_to_char_suffix( filename_head, it )
        open(1, file = filename)
           do i = 1, N
              do j = 1, M
                 write(1,*) i, j, real( phi(i, j) )
              enddo
	      write(1,'(a)') '#' 		!this line is for gnuplot
           enddo
        close(1)
	   		
	end subroutine output

!==========================================================================

	END MODULE poisson_routines
