      	program poisson


!===================================================================================
!  14 Apr 98
!  Paul M. Alsing
!  Albuqueruqe High Performance Computing Center
!  1601 Central NE
!  University of New Mexico
!  Albuquerque, New Mexico 87131
!  alsing@arc.unm.edu
!  505-277-9094
!===================================================================================
! Iteratively solve Poisson Equation 
!             d^2phi(x,y)/dx^2 + d^2phi(x,y)/dy^2 = -4*pi*rho(x,y)
! for the potential phi(x,y) given the charge density rho(x,y) on a N^2 point
! grid of physical length L on each side.  Use the (1) F90 routines cshift() or 
! (2) modulo arithmetic indexing, to take into account periodic boundary conditions.
!===================================================================================


	use poisson_routines
      	use string_utilities

      	implicit none
      

      	integer 			:: N, it
	integer				:: i,j
	integer				:: method

      	real(DP) 			:: L,dx,norm,rn,kx2,ky2
      	real(DP), allocatable 		:: rho(:,:), phi(:,:), phi_old(:,:)      
      	real(DP), parameter 		:: eps = 1d-5
   	real(DP)			:: rtc ! for AIX only
        real(dp),external               :: cpu_time
	real(DP)			:: t_start, t_end
 
! User inputs

      	print*, 'Input the length of each side of the grid L, and the # of grid points N'
      	read*, L,N
	print*,'Solving Poisson Equation for L = ',real(L),'  N = ',N

	print*,'Input the method of computation'
	print*,'1 -> F90 cshifts'
	print*,'2 -> modulo indexing'
	read*, method
	if(method==1) print*,'Using F90 cshifts'
	if(method==2) print*,'Using modulo indexing'
 

!  allocate memory for arrays
     
      	allocate (rho(N,N))		!charge density
      	allocate (phi(N,N))		!potential
      	allocate (phi_old(N,N))		!potential, previous iteration

! initialization

      	dx=L/N
      	rho = makerho(L,N)*pi*dx*dx	!initial charge density

! diagnostic: output the charge density rho 

        open(1,file='rho_serial.bin',form='unformatted',access='direct',recl=4*N*N)
        write(1,rec=1) real( rho/(pi*dx*dx) )                      
        close(1)


      	phi=0.0				!set potential to 0
      	rn = sum(abs(rho))		!sum of abs(rho) over whole grid
!      	norm=.1


!  Iteratively solve Poisson Equation

	it = 0

  	t_start = cpu_time()  ! rtc() - for AIX only

	do 

	   it = it + 1

           phi_old=phi		!save old value of phi

	   ! one iteration: cshifts take into account periodic boudnary condition

	   select case( method )

	    case( 1 )		! using cshifts

               phi = rho + 1/4.*(cshift(phi, 1, 1) + cshift(phi, -1, 1) +     &
                                 cshift(phi, 1, 2) + cshift(phi, -1, 2) )
	    case( 2 )

	       do j = 1, N
	          do  i = 1, N
	              phi(i,j) = rho(i,j) 						&
		      	       + 0.25_dp * (   phi_old( modulo(i,N)  +1, j )	&	!u(i+1,j)		
		   	       		    +  phi_old( modulo(i-2,N)+1, j )	&	!u(i-1,j)
			      		    +  phi_old( i, modulo(j,N)+1   )	&	!u(i,j+1)
					    +  phi_old( i, modulo(j-2,N)+1 )   	&	!u(i,j-1)
				           )
	          enddo
	       enddo

	    end select

	   ! The solution phi is undetermined up to an arbitray constant.
	   ! We fix this constant by subtracting off, at each iteration,
	   ! the average value of phi over the whole grid 

           phi=phi - sum(phi)/N/N	!sum is over both dimensions of phi	 


	   ! compute norm of difference between current and old solution

           norm = sum(abs(phi-phi_old))/rn

	   if(norm < eps) exit


	   ! test for non-convergence

	   if( it >= ITMAX ) then
	       print*,'*** Solution failed to converger after ',ITMAX,' iterations'
	       stop 'EXITING'
	   endif

	   ! periodic output

	   if( modulo( it, NOUTPUT ) == 0 ) then
	       call output( phi, filename_head, it)
	       print*,'  At it = ',it,' norm = ',real(norm)
	   endif

	enddo


!  final output

	print*,'At it = ',it,' norm = ',real(norm),'  comvergence!'
  	t_end = cpu_time() ! rtc() - for AIX only
	print*,'Execution time = ',t_end-t_start,' secs for ',it,' iterations'
	print*,'or ',(t_end-t_start)/it,' secs per iteration. N = ',N
	call output( phi, filename_head, it)

!  end of program
         
	print'(/,a)','END OF PROGRAM'
      	end


